#pragma once
#include <SFML/Graphics.hpp>
#include <map>
#include <string>
#include <fstream>
#include <math.h>
#include <memory>
#include <random>
namespace zle
{
	static const std::string ZLE_VERSION = "#1.4";
	class TileMap : public sf::Transformable, public sf::Drawable
	{
		mutable sf::VertexArray arr;
		mutable sf::VertexBuffer buff;
		sf::Vector2u mapSize;
		sf::Vector2u tileSize;
		sf::Vector2u textureTileSize;
		sf::Color overAllColor;
		sf::Texture* texture;
		unsigned int textureGap;
		sf::Color mapColor;
		mutable bool updateBeforeDraw;
		std::vector<std::vector<unsigned short>> mapData;
		void InternalUpdate()
		{
			//resize buffer
			mapData.resize(mapSize.x);
			for (int i = 0; i < mapSize.x; i++)
				mapData[i].resize(mapSize.y);
		}
		void DrawUpdate() const
		{
			updateBeforeDraw = 0;
			arr.setPrimitiveType(sf::Triangles);
			arr.clear();
			arr.resize(6 * mapSize.x * mapSize.y);
			if (sf::VertexBuffer::isAvailable())
			{
				buff.setPrimitiveType(sf::Triangles);
				buff.setUsage(sf::VertexBuffer::Dynamic);
			}
			//setup data
			for (int i = 0; i < mapSize.x; i++)
				for (int j = 0; j < mapSize.y; j++)
				{
					sf::Vector2f position = sf::Vector2f(tileSize.x * i, tileSize.y * j);
					unsigned int v = (j + mapSize.y * i) * 6;

					arr[v].position = position;
					arr[v].color = overAllColor;
					arr[v].texCoords = sf::Vector2f(mapData[i][j] * textureTileSize.x + textureGap * mapData[i][j], 0);

					arr[v + 1].position = position + sf::Vector2f(tileSize.x, 0);
					arr[v + 1].color = overAllColor;
					arr[v + 1].texCoords = sf::Vector2f(mapData[i][j] * textureTileSize.x + textureTileSize.x + textureGap * mapData[i][j], 0);

					arr[v + 2].position = position + sf::Vector2f(tileSize);
					arr[v + 2].color = overAllColor;
					arr[v + 2].texCoords = sf::Vector2f(mapData[i][j] * textureTileSize.x + textureTileSize.x + textureGap * mapData[i][j], textureTileSize.y);

					arr[v + 3].position = position + sf::Vector2f(0, tileSize.y);
					arr[v + 3].color = overAllColor;
					arr[v + 3].texCoords = sf::Vector2f(mapData[i][j] * textureTileSize.x + textureGap * mapData[i][j], textureTileSize.y);

					arr[v + 4] = arr[v];
					arr[v + 5] = arr[v + 2];
				}

			//update
			if (sf::VertexBuffer::isAvailable() && arr.getVertexCount() > 0)
			{
				if (buff.getVertexCount() != arr.getVertexCount())
					buff.create(arr.getVertexCount());
				buff.update(&arr[0]);
			}
		}
		template<typename T>
		static void Save(std::string& s, T var)
		{
			for (int i = 0; i < sizeof(T); i++)
				s += (reinterpret_cast<char*>(&var))[i];
		}
		template<typename T>
		static void Save(std::ofstream& s, T var)
		{
			for (int i = 0; i < sizeof(T); i++)
				s << (reinterpret_cast<char*>(&var))[i];
		}
		static void SaveStr(std::ofstream& s, const std::string& var)
		{
			sf::Uint32 size = var.size();
			for (int i = 0; i < sizeof(sf::Uint32); i++)
				s << (reinterpret_cast<char*>(&size))[i];
			s << var;
		}
		template<typename T>
		static void Load(sf::InputStream& input, T& t)
		{
			input.read(&t, sizeof(T));
		}
		static void LoadStr(sf::InputStream& input, std::string& data)
		{
			sf::Uint32 size;
			input.read(&size, sizeof(sf::Uint32));
			data.resize(size);
			input.read(&data[0], size);
		}
	public:
		TileMap()
			: texture(nullptr), textureGap(0), updateBeforeDraw(0)
		{
			overAllColor = sf::Color::White;
			mapSize = sf::Vector2u(1, 1);
			tileSize = sf::Vector2u(1, 1);
			textureTileSize = sf::Vector2u(1, 1);
		}
		~TileMap() {}

		void saveToFile(const std::string& fileName)
		{
			std::ofstream save1;
			save1.open(fileName, std::ios::binary);
			if (save1.is_open())
			{
				Save(save1, getTileCount().x);
				Save(save1, getTileCount().y);
				for (int i = 0; i < getTileCount().x; i++)
					for (int j = 0; j < getTileCount().y; j++)
					{
						Save(save1, getTileID(sf::Vector2u(i, j)));
					}
				save1.close();
			}
		}
		void saveToMemory(std::string& memory)
		{
			memory = "";
			Save(memory, getTileCount().x);
			Save(memory, getTileCount().y);
			for (int i = 0; i < getTileCount().x; i++)
				for (int j = 0; j < getTileCount().y; j++)
				{
					Save(memory, getTileID(sf::Vector2u(i, j)));
				}
		}
		void loadFromMemory(const char* data, sf::Uint64 Size)
		{
			sf::MemoryInputStream stream;
			stream.open(data, Size);
			sf::Vector2u size;
			Load(stream, size.x);
			Load(stream, size.y);

			setTileCount(size);
			for (int i = 0; i < getTileCount().x; i++)
				for (int j = 0; j < getTileCount().y; j++)
				{
					unsigned short var;
					Load(stream, var);
					setTileID(sf::Vector2u(i, j), var);
				}
		}
		void loadFromStream(sf::FileInputStream& stream)
		{
			sf::Vector2u size;
			Load(stream, size.x);
			Load(stream, size.y);
			setTileCount(size);
			for (int i = 0; i < getTileCount().x; i++)
				for (int j = 0; j < getTileCount().y; j++)
				{
					unsigned short var;
					Load(stream, var);
					setTileID(sf::Vector2u(i, j), var);
				}
		}
		void loadFromFile(const std::string& fileName)
		{
			sf::FileInputStream load1;
			if (load1.open(fileName))
			{
				loadFromStream(load1);
			}
		}

		void setTileCount(const sf::Vector2u& mapSize)
		{
			this->mapSize = mapSize;
			InternalUpdate();
		}
		const sf::Vector2u& getTileCount() const
		{
			return mapSize;
		}
		void setTexture(sf::Texture* texture)
		{
			this->texture = texture;
		}
		const sf::Texture& getTexture() const
		{
			return *texture;
		}
		void setTileSize(const sf::Vector2u& tileSize)
		{
			this->tileSize = tileSize;
			updateBeforeDraw = 1;
		}
		const sf::Vector2u& getTileSize() const
		{
			return tileSize;
		}
		void setTextureTileSize(const sf::Vector2u& textureTileSize)
		{
			this->textureTileSize = textureTileSize;
			updateBeforeDraw = 1;
		}
		void setTextureGap(unsigned int gap)
		{
			textureGap = gap;
			updateBeforeDraw = 1;
		}
		unsigned int getTextureGap() const
		{
			return textureGap;
		}
		sf::Vector2u getTextureTileSize() const
		{
			return textureTileSize;
		}
		void setColor(sf::Color color)
		{
			overAllColor = color;
			updateBeforeDraw = 1;
		}
		sf::Color getColor() const
		{
			return overAllColor;
		}
		unsigned short getTileID(sf::Vector2u tile) const
		{
			return mapData[tile.x][tile.y];
		}
		void setTileID(sf::Vector2u tile, unsigned short ID)
		{
			mapData[tile.x][tile.y] = ID;
			updateBeforeDraw = 1;
		}
	private:
		virtual void draw(sf::RenderTarget& target, sf::RenderStates states) const
		{
			states.transform *= getTransform();
			states.texture = texture;
			if (updateBeforeDraw)
				DrawUpdate();
			if (sf::VertexBuffer::isAvailable())
				target.draw(buff, states);
			else
				target.draw(arr, states);
		}
	};
	class ParticleSystemEvent
	{
	public:
		enum Type
		{
			OnUpdate = 1,
			OnDeath = 1 << 1,
			OnCreate = 1 << 2,
			All = OnUpdate | OnDeath | OnCreate,
			Count = 3
		};
		Type type = Type::All;
		sf::Vector2f particlePosition = sf::Vector2f();
		float particleRotation = 0;
		sf::Color particleColor = sf::Color();
		sf::Time lifeTime = sf::Time();
		ParticleSystemEvent()
		{

		}
		ParticleSystemEvent(const Type& type, const sf::Vector2f& particlePosition, const float& particleRotation, const sf::Color& particleColor, const sf::Time& lifeTime)
			: type(type), particlePosition(particlePosition), particleRotation(particleRotation), particleColor(particleColor), lifeTime(lifeTime)
		{
		}
		~ParticleSystemEvent()
		{

		}
	};
	class ParticleSystem : public sf::Drawable, public sf::Transformable
	{
	public:
		enum class ParticleType;
	private:
		struct Particle : public sf::Transformable
		{
			sf::Uint8 vertexCount;
			sf::Vertex* vertices;
			sf::Color currentColor;
			sf::Color startColor;
			sf::Color endColor;
			sf::Time lifeTime;
			sf::Time totalLifeTime;
			sf::Vector2f direction;
			sf::Vector2f offset;
			sf::Vector2f force;
			float startSize;
			float endSize;
			float startSpeed;
			float endSpeed;
			bool inUse;
			Particle()
				:inUse(0), offset(0, 0), direction(0, 0)
			{

			}
			void resetPointCoord()
			{
				switch (vertexCount)
				{
				case 1:
					vertices[0].texCoords = sf::Vector2f(0, 0);
				case 2: case 3:
				{
					float add = 360.0 / vertexCount;
					for (int i = 0; i < vertexCount; i++)
					{
						vertices[i].texCoords = sf::Vector2f(cos((add * i - 45) * 0.0174533), sin((add * i - 45) * 0.0174533));
					}
				}
				break;
				case 4:
					vertices[0].texCoords = sf::Vector2f(-1, -1);
					vertices[1].texCoords = sf::Vector2f(1, -1);
					vertices[2].texCoords = sf::Vector2f(1, 1);
					vertices[3].texCoords = sf::Vector2f(-1, 1);
					break;
				case 6:
					vertices[0].texCoords = sf::Vector2f(-1, -1);
					vertices[1].texCoords = sf::Vector2f(1, -1);
					vertices[2].texCoords = sf::Vector2f(1, 1);
					vertices[3].texCoords = sf::Vector2f(-1, 1);
					vertices[4].texCoords = vertices[0].texCoords;
					vertices[5].texCoords = vertices[2].texCoords;
					break;
				default:
					vertices[0].texCoords = sf::Vector2f(0, 0);
					break;
				}
			}
			const sf::Vector2f getPoint(int index)
			{
				switch (vertexCount)
				{
				case 1:
					return sf::Vector2f(0, 0);
				case 2: case 3: case 4:
				{
					float add = 360.0 / vertexCount;
					return sf::Vector2f(cos((add * index - 90) * 0.0174533), sin((add * index - 90) * 0.0174533));
					break;
				}
				case 6:
				{
					float add = 360.0 / 4;
					if (index == 4)
						return sf::Vector2f(cos((-90) * 0.0174533), sin((-90) * 0.0174533));
					if (index == 5)
						return sf::Vector2f(cos((add * 2 - 90) * 0.0174533), sin((add * 2 - 90) * 0.0174533));
					return sf::Vector2f(cos((add * index - 90) * 0.0174533), sin((add * index - 90) * 0.0174533));
					break;
				}
				}
			}
			void Update()
			{
				if (vertexCount == 1)
				{
					vertices[0].position = getPosition();
					vertices[0].color = currentColor;
				}
				else
					for (int i = 0; i < vertexCount; i++)
					{
						vertices[i].position = getTransform().transformPoint(getPoint(i));
						vertices[i].color = currentColor;
					}


			}
		};
		//event items
		void* spawnOnEvent[ParticleSystemEvent::Count];
		bool inherit[3][ParticleSystemEvent::Count];

		//random items
		std::mt19937 randomFunc;
		float r_lifeTime;
		float r_startSpeed;
		float r_endSpeed;
		float r_startSize;
		float r_endSize;
		sf::Color r_secondaryStart;
		sf::Color r_secondaryEnd;
		bool r_useSecondary;

		mutable std::shared_ptr<sf::Mutex> updated;
		mutable std::shared_ptr<sf::Mutex> updated2;
		std::shared_ptr<sf::Thread> thread;
		std::vector<ParticleSystemEvent> events;
		std::vector<sf::Vertex> arr;
		sf::VertexBuffer buff;
		std::vector<Particle> particles;
		sf::Time updateTime = sf::Time::Zero;
		sf::Time cntDown = sf::Time::Zero;
		mutable unsigned int totalParticles = 0;
		sf::Uint8 eventFlags = 0;

		int updateCalls = 0;
		//every particle variable
		sf::Vector2f startPos = sf::Vector2f();
		unsigned int maxParticles;
		float lifeSeconds;
		ParticleType type;
		float startRotation;
		float endRotation;
		unsigned int spawnCount;
		float spawnSeconds;
		float spawnRadius;
		sf::Vector2f constantForce;
		sf::Vector2f startForce;
		float startSize;
		float endSize;
		float startSpeed;
		float endSpeed;
		float fireAngle;
		float fireRotation;
		sf::Color startColor;
		sf::Color endColor;
		float fading;
		bool randomStartRotation;
		bool createOnUpdate;
		const sf::Texture* texture;
		bool firstUpdate;
		bool keepUpWithFrameRate;
		bool drawNewestOnTop;
		bool useThread;
		bool updateSystem;
		bool returned = 0;
		sf::Time createFor = sf::Time::Zero;

		unsigned int verticesPerParticle() const
		{
			switch (type)
			{
			case ParticleSystem::ParticleType::Points:
				return 1;
				break;
			case ParticleSystem::ParticleType::Lines:
				return 2;
				break;
			case ParticleSystem::ParticleType::Triangles:
				return 3;
				break;
			case ParticleSystem::ParticleType::Quads:
				return 4;
				break;
			case ParticleSystem::ParticleType::QuadsTriangles:
				return 6;
				break;
			default:
				return 1;
				break;
			}
		}
		float getParam(const sf::Time& lifeTotal, const sf::Time& lifeLeft, float start, float end) const
		{
			float zeroToOne = (lifeTotal.asSeconds() - lifeLeft.asSeconds()) / lifeTotal.asSeconds();
			return (end - start) * zeroToOne + start;
		}
		void refreshPointer()
		{
			if (!drawNewestOnTop)
				return;
			for (int j = 0; j < particles.size(); j++)
			{
				particles[j].vertices = &arr[j * verticesPerParticle()];
			}
		}
		float getRandomValueTrig()
		{
			return static_cast<float>(randomFunc()) / randomFunc.max() * 6.283184f;
		}
		float getRandomValue01()
		{
			return static_cast<float>(randomFunc()) / randomFunc.max();
		}
		float getRandomValue11()
		{
			return static_cast<float>(randomFunc()) / (randomFunc.max() / 2) - 1;
		}
		int findFreeParticle()
		{
			for (int i = 0; i < particles.size(); i++)
			{
				if (!particles[i].inUse)
					return i;
			}
			return -1;
		}
		void setupArray()
		{
			sf::Lock lock2(*updated2);
			deleteParticles();
			if (drawNewestOnTop)
				return;
			particles.clear();
			arr.clear();
			totalParticles = 0;
			particles.resize(maxParticles);
			arr.resize(maxParticles * verticesPerParticle());
			for (int i = 0; i < particles.size(); i++)
			{
				particles[i].vertexCount = verticesPerParticle();
				particles[i].vertices = &arr[i * verticesPerParticle()];
			}

		}
		void applyTexture(const sf::Texture* newTexture)
		{
			if (newTexture == nullptr)
				return;
			sf::Vector2f size = static_cast<sf::Vector2f>(newTexture->getSize());
			if (type == ParticleType::Points)
			{
				for (int i = 0; i < particles.size(); i++)
				{
					particles[i].vertices[0].texCoords = sf::Vector2f(size.x / 2, size.y / 2);
				}
			}
			else
				for (int i = 0; i < particles.size(); i++)
				{
					particles[i].resetPointCoord();
					for (int j = 0; j < particles[i].vertexCount; j++)
					{
						sf::Vector2f& v = particles[i].vertices[j].texCoords;
						v = sf::Vector2f((v.x + 1) / 2.0 * size.x, (v.y + 1) / 2.0 * size.y);
					}
				}

		}
		void ParticleEvents(ParticleSystemEvent evnt, ParticleSystem* target, int ID)
		{
			ParticleSystem& o = *reinterpret_cast<ParticleSystem*>(target->spawnOnEvent[ID]);
			sf::Vector2f startPos = o.getSpawnPosition();
			sf::Color startColor = o.getStartColor();
			float lifeTime = o.getLifeTime();
			if (target->inherit[0][ID])
				o.setSpawnPosition(evnt.particlePosition);
			if (target->inherit[1][ID])
				o.setStartColor(evnt.particleColor);
			if (target->inherit[2][ID])
				o.setLifeTime(evnt.lifeTime.asSeconds());
			o.Create();
			if (target->inherit[0][ID])
				o.setSpawnPosition(startPos);
			if (target->inherit[1][ID])
				o.setStartColor(startColor);
			if (target->inherit[2][ID])
				o.setLifeTime(lifeTime);
		}
		void threadLoop()
		{
			sf::Clock clock;
			returned = 0;
			while (1)
			{
				sf::Clock clock2;
				sf::Time t = clock.restart();
				if (updateSystem)
				{
					Update(t);
					ParticleSystemEvent evnt;
					while (pollEvent(evnt))
					{
						if (evnt.type == ParticleSystemEvent::OnCreate && spawnOnEvent[0])
						{
							ParticleEvents(evnt, this, 0);
						}
						if (evnt.type == ParticleSystemEvent::OnUpdate && spawnOnEvent[1])
						{
							ParticleEvents(evnt, this, 1);
						}
						if (evnt.type == ParticleSystemEvent::OnDeath && spawnOnEvent[2])
						{
							ParticleEvents(evnt, this, 2);
						}
					}
				}
				if (!useThread)
				{
					returned = 1;
					return;
				}
				if (clock2.getElapsedTime().asMicroseconds() < 16667)
					sf::sleep(sf::microseconds(16667 - clock2.getElapsedTime().asMicroseconds()));
			}
			return;
		}
		void createEvent(const ParticleSystemEvent::Type& type, const Particle& particle)
		{
			if ((eventFlags & (int)type) > 0)
				events.emplace_back(type, particle.getPosition(), particle.getRotation(), particle.currentColor, particle.lifeTime);
		}

		void setup()
		{
			thread = std::make_shared<sf::Thread>(&ParticleSystem::threadLoop, this);
			updated = std::make_shared<sf::Mutex>();
			updated2 = std::make_shared<sf::Mutex>();
			buff.setUsage(sf::VertexBuffer::Stream);
			if (type == ParticleType::QuadsTriangles)
				buff.setPrimitiveType(sf::PrimitiveType::Triangles);
			else
				buff.setPrimitiveType(static_cast<sf::PrimitiveType>(type));
			setupArray();
			useSeparateThread(useThread);
		}
		template<typename T>
		static void Save(std::string& s, T var)
		{
			for (int i = 0; i < sizeof(T); i++)
				s += (reinterpret_cast<char*>(&var))[i];
		}
		template<typename T>
		static void Save(std::ofstream& s, T var)
		{
			for (int i = 0; i < sizeof(T); i++)
				s << (reinterpret_cast<char*>(&var))[i];
		}
		static void SaveStr(std::ofstream& s, const std::string& var)
		{
			sf::Uint32 size = var.size();
			for (int i = 0; i < sizeof(sf::Uint32); i++)
				s << (reinterpret_cast<char*>(&size))[i];
			s << var;
		}
		template<typename T>
		static void Load(sf::InputStream& input, T& t)
		{
			input.read(&t, sizeof(T));
		}
		static void LoadStr(sf::InputStream& input, std::string& data)
		{
			sf::Uint32 size;
			input.read(&size, sizeof(sf::Uint32));
			data.resize(size);
			input.read(&data[0], size);
		}
	public:
		enum class ParticleType
		{
			Points = sf::Points,
			Lines = sf::Lines,
			LineStrip = sf::LineStrip,
			Triangles = sf::Triangles,
			TriangleStrip = sf::TriangleStrip,
			TriangleFan = sf::TriangleFan,
			Quads = sf::Quads,
			QuadsTriangles
		};

		ParticleSystem()
			: maxParticles(100), lifeSeconds(1), type(ParticleType::Triangles), startRotation(0), endRotation(0), useThread(0),
			spawnCount(1), spawnSeconds(0.1), constantForce(0, 0), startForce(0, 0), startSize(2), endSize(2),
			startSpeed(10), endSpeed(10), startColor(255, 255, 255, 255), endColor(255, 255, 255, 255), fading(1),
			randomStartRotation(0), createOnUpdate(1), texture(nullptr), firstUpdate(0), keepUpWithFrameRate(0),
			drawNewestOnTop(1), fireAngle(360), fireRotation(0), updateSystem(1), spawnRadius(0), r_lifeTime(0),
			r_endSize(0), r_startSize(0), r_startSpeed(0), r_endSpeed(0), r_useSecondary(0), r_secondaryEnd(255, 255, 255, 255), r_secondaryStart(255, 255, 255, 255),
			randomFunc(time(NULL))
		{
			for (int i = 0; i < ParticleSystemEvent::Count; i++)
			{
				for (int j = 0; j < 3; j++)
					inherit[j][i] = 0;
				spawnOnEvent[i] = nullptr;
			}
			setup();
		}
		~ParticleSystem()
		{
			if (useThread)
			{
				useThread = 0;
				sf::Clock clock;
				while (!returned)
				{
					if (clock.getElapsedTime().asSeconds() > 1)
						break;
					sf::sleep(sf::milliseconds(1));
				}
			}
			thread->terminate();
		}

		void saveToFile(const std::string& fileName) const
		{
			std::ofstream save1;
			save1.open(fileName, std::ios::binary);
			saveToFile(save1);
			save1.close();
		}
		void saveToFile(std::ofstream& save1) const
		{
			if (save1.is_open())
			{
				std::string data = "";
				saveToMemory(data);
				save1 << data;
			}
		}
		void saveToMemory(std::string& save1) const
		{
			Save(save1, getConstantForce());
			Save(save1, getEndColor());
			Save(save1, getEndRotation());
			Save(save1, getEndSize());
			Save(save1, getEndSpeed());
			Save(save1, getFading());
			Save(save1, getLifeTime());
			Save(save1, getMaxParticles());
			Save(save1, getParticleType());
			Save(save1, getRandomStartRotation());
			Save(save1, getSpawnAngle());
			Save(save1, getSpawnCount());
			Save(save1, getSpawnRate());
			Save(save1, getSpawnRotation());
			Save(save1, getStartColor());
			Save(save1, getStartForce());
			Save(save1, getStartRotation());
			Save(save1, getStartSize());
			Save(save1, getStartSpeed());
			Save(save1, getDrawNewestOnTop());
			Save(save1, getKeepUpWithFrameRate());
			Save(save1, getSpawnPosition());
			Save(save1, r_lifeTime);
			Save(save1, r_secondaryEnd);
			Save(save1, r_secondaryStart);
			Save(save1, r_endSize);
			Save(save1, r_startSize);
			Save(save1, r_endSpeed);
			Save(save1, r_startSpeed);
			Save(save1, getSpawnRadius());
		}
		void loadFromStream(sf::InputStream& file)
		{
			Load(file, constantForce);
			Load(file, endColor);
			Load(file, endRotation);
			Load(file, endSize);
			Load(file, endSpeed);
			Load(file, fading);
			Load(file, lifeSeconds);
			Load(file, maxParticles);
			Load(file, type);
			Load(file, randomStartRotation);
			Load(file, fireAngle);
			Load(file, spawnCount);
			Load(file, spawnSeconds);
			Load(file, fireRotation);
			Load(file, startColor);
			Load(file, startForce);
			Load(file, startRotation);
			Load(file, startSize);
			Load(file, startSpeed);
			Load(file, drawNewestOnTop);
			Load(file, keepUpWithFrameRate);
			Load(file, startPos);
			Load(file, r_lifeTime);
			Load(file, r_secondaryEnd);
			Load(file, r_secondaryStart);
			Load(file, r_endSize);
			Load(file, r_startSize);
			Load(file, r_endSpeed);
			Load(file, r_startSpeed);
			Load(file, spawnRadius);
			setParticleType(type);
		}
		void loadFromFile(const std::string& fileName)
		{
			sf::FileInputStream load1;
			if (load1.open(fileName))
			{
				loadFromStream(load1);
			}
		}

		const float& getSpawnAngle() const
		{
			return fireAngle;
		}
		const float& getSpawnRotation() const
		{
			return fireRotation;
		}
		const unsigned int& getMaxParticles() const
		{
			return maxParticles;
		}
		const float& getLifeTime() const
		{
			return lifeSeconds;
		}
		const ParticleType& getParticleType() const
		{
			return type;
		}
		const float& getStartRotation() const
		{
			return startRotation;
		}
		const float& getEndRotation() const
		{
			return endRotation;
		}
		const unsigned int& getSpawnCount() const
		{
			return spawnCount;
		}
		const float& getSpawnRate() const
		{
			return spawnSeconds;
		}
		const sf::Vector2f& getConstantForce() const
		{
			return constantForce;
		}
		const sf::Vector2f& getStartForce() const
		{
			return startForce;
		}
		const float& getStartSize() const
		{
			return startSize;
		}
		const float& getEndSize() const
		{
			return endSize;
		}
		const float& getStartSpeed() const
		{
			return startSpeed;
		}
		const float& getEndSpeed() const
		{
			return endSpeed;
		}
		const sf::Color& getStartColor() const
		{
			return startColor;
		}
		const sf::Color& getEndColor() const
		{
			return endColor;
		}
		const float& getFading() const
		{
			return fading;
		}
		const bool& getRandomStartRotation() const
		{
			return randomStartRotation;
		}
		const sf::Texture& getTexture() const
		{
			return *texture;
		}
		const unsigned int& getActiveParticles() const
		{
			if (drawNewestOnTop)
				totalParticles = particles.size();
			return totalParticles;
		}
		const sf::Vector2f& getSpawnPosition() const
		{
			return startPos;
		}
		const float& getSpawnRadius() const
		{
			return spawnRadius;
		}
		const float& getRandomLifeTime() const
		{
			return r_lifeTime;
		}
		const float& getRandomStartSpeed() const
		{
			return r_startSpeed;
		}
		const float& getRandomEndSpeed() const
		{
			return r_endSpeed;
		}
		const float& getRandomStartSize() const
		{
			return r_startSize;
		}
		const float& getRandomEndSize() const
		{
			return r_endSize;
		}
		const sf::Color& getSecondStartColor() const
		{
			return r_secondaryStart;
		}
		const sf::Color& getSecondEndColor() const
		{
			return r_secondaryEnd;
		}

		void setCaptureEvent(bool state, sf::Uint8 type = (sf::Uint8)ParticleSystemEvent::All)
		{
			if (state)
			{
				eventFlags |= type;
			}
			else
			{
				eventFlags &= ~type;
			}
		}
		void deleteParticles()
		{
			sf::Lock lock2(*updated2);
			particles.clear();
			arr.clear();
			totalParticles = 0;
		}
		bool getDrawNewestOnTop() const
		{
			return drawNewestOnTop;
		}
		bool getKeepUpWithFrameRate() const
		{
			return keepUpWithFrameRate;
		}
		bool pollEvent(ParticleSystemEvent& event)
		{
			if (events.size() > 0)
			{
				event = events.back();
				events.pop_back();
				return 1;
			}
			return 0;
		}
		const std::vector<ParticleSystemEvent>& getEvents() const
		{
			return events;
		}
		void deleteEvents()
		{
			events.clear();
		}
		void useSeparateThread(bool state)
		{
			useThread = state;
			if (useThread)
				thread->launch();
		}
		void setDrawNewestOnTop(bool state)
		{
			drawNewestOnTop = state;
			setupArray();
			applyTexture(texture);
		}
		void setSpawnAngle(float fireAngle)
		{
			this->fireAngle = fireAngle;
		}
		void setSpawnPosition(const sf::Vector2f& pos)
		{
			this->startPos = pos;
		}
		void setSpawnRotation(float rotation)
		{
			this->fireRotation = rotation;
		}
		void setMaxParticles(unsigned int count)
		{
			maxParticles = count;
			setupArray();
			applyTexture(texture);
		}
		void setLifeTime(float seconds)
		{
			lifeSeconds = seconds;
		}
		void setRandomLifeTime(float seconds)
		{
			r_lifeTime = seconds;
		}
		void setRandomStartSpeed(float speed)
		{
			r_startSpeed = speed;
		}
		void setRandomEndSpeed(float speed)
		{
			r_endSpeed = speed;
		}
		void setRandomStartSize(float size)
		{
			r_startSize = size;
		}
		void setRandomEndSize(float size)
		{
			r_endSize = size;
		}
		void useSecondColors(bool state)
		{
			r_useSecondary = state;
		}
		void setSecondStartColor(const sf::Color& color)
		{
			r_secondaryStart = color;
		}
		void setSecondEndColor(const sf::Color& color)
		{
			r_secondaryEnd = color;
		}
		void setParticleType(ParticleType type)
		{
			sf::Lock lock(*updated);
			sf::Lock lock2(*updated2);
			this->type = type;
			if (type == ParticleType::QuadsTriangles)
				buff.setPrimitiveType(sf::PrimitiveType::Triangles);
			else
				buff.setPrimitiveType((sf::PrimitiveType)type);
			setupArray();
			applyTexture(texture);

		}
		void setStartRotation(float rotation)
		{
			startRotation = rotation;
		}
		void setEndRotation(float rotation)
		{
			endRotation = rotation;
		}
		void setSpawnCount(unsigned int count)
		{
			spawnCount = count;
		}
		void setSpawnRate(float time)
		{
			spawnSeconds = time;
			cntDown = sf::Time::Zero;
			createFor = sf::Time::Zero;
		}
		void setConstantForce(sf::Vector2f force)
		{
			constantForce = force;
		}
		void setStartForce(sf::Vector2f force)
		{
			startForce = force;
		}
		void setStartSize(float size)
		{
			startSize = size;
		}
		void setEndSize(float size)
		{
			endSize = size;
		}
		void setStartSpeed(float speed)
		{
			startSpeed = speed;
		}
		void setEndSpeed(float speed)
		{
			endSpeed = speed;
		}
		void setStartColor(const sf::Color& color)
		{
			startColor = color;
		}
		void setEndColor(const sf::Color& color)
		{
			endColor = color;
		}
		void setFading(float percentage)
		{
			fading = percentage;
		}
		void setRandomStartRotation(bool state)
		{
			randomStartRotation = state;
		}
		void setTexture(const sf::Texture& texture)
		{
			applyTexture(&texture);
			this->texture = &texture;
		}
		void setCreateOnUpdate(bool createOnUpdate)
		{
			this->createOnUpdate = createOnUpdate;
		}
		void setKeepUpWithFrameRate(bool state)
		{
			keepUpWithFrameRate = state;
			cntDown = sf::Time::Zero;
			createFor = sf::Time::Zero;
		}
		void setSpawnRadius(float radius)
		{
			spawnRadius = radius;
		}
		void Clear()
		{
			for (int i = 0; i < particles.size(); i++)
			{
				if (particles[i].inUse)
					particles[i].lifeTime = sf::Time::Zero;
			}
			Update(sf::Time::Zero);
		}
		void CreateFor(const sf::Time& time)
		{
			createFor = time;
			Create();
		}
		void Create()
		{
			if (drawNewestOnTop)
				for (int j = 0; j < spawnCount; j++)
				{
					if (arr.size() / verticesPerParticle() >= maxParticles)
						return;
					Particle particle;
					particle.vertexCount = verticesPerParticle();
					for (int i = 0; i < verticesPerParticle(); i++)
						arr.push_back(sf::Vertex());
					particle.vertices = &arr[arr.size() - verticesPerParticle()];
					particle.lifeTime = sf::seconds(lifeSeconds + getRandomValue11() * r_lifeTime);
					particle.totalLifeTime = particle.lifeTime;
					sf::Color startRandom;
					if (r_useSecondary)
						startRandom = sf::Color(startColor.r + getRandomValue01() * (r_secondaryStart.r - startColor.r),
							startColor.g + getRandomValue01() * (r_secondaryStart.g - startColor.g),
							startColor.b + getRandomValue01() * (r_secondaryStart.b - startColor.b));
					else
						startRandom = startColor;

					sf::Color endRandom;
					if (r_useSecondary)
						endRandom = sf::Color(endColor.r + getRandomValue01() * (r_secondaryEnd.r - endColor.r),
							endColor.g + getRandomValue01() * (r_secondaryEnd.g - endColor.g),
							endColor.b + getRandomValue01() * (r_secondaryEnd.b - endColor.b));
					else
						endRandom = endColor;
					particle.currentColor = startRandom;
					particle.startColor = startRandom;
					particle.endColor = endRandom;
					particle.force = startForce;
					particle.inUse = 1;
					particle.startSize = startSize + getRandomValue11() * r_startSize;
					particle.endSize = endSize + getRandomValue11() * r_endSize;
					particle.startSpeed = startSpeed + getRandomValue11() * r_startSpeed;
					particle.endSpeed = endSpeed + getRandomValue11() * r_endSpeed;
					float randomDirection = getRandomValue01() * fireAngle + fireRotation;
					particle.direction = sf::Vector2f(cos(randomDirection * 0.0174533), sin(randomDirection * 0.0174533));
					particle.setRotation(randomStartRotation ? getRandomValue01() * 360 : 0);
					if (texture != nullptr)
					{
						sf::Vector2f size = (sf::Vector2f)texture->getSize();
						particle.resetPointCoord();
						for (int j = 0; j < particle.vertexCount; j++)
						{
							sf::Vector2f& v = particle.vertices[j].texCoords;
							v = sf::Vector2f((v.x + 1) / 2.0 * size.x, (v.y + 1) / 2.0 * size.y);
						}
					}
					float random1 = getRandomValueTrig();
					float random2 = getRandomValue01();
					sf::Vector2f randomPos = startPos + sf::Vector2f(
						cos(random1) * random2 * spawnRadius,
						sin(random1) * random2 * spawnRadius);
					particle.offset = randomPos;
					particle.setPosition(randomPos);
					totalParticles++;
					particle.Update();
					particles.push_back(particle);
					createEvent(ParticleSystemEvent::Type::OnCreate, particle);
				}
			else
			{
				for (int j = 0; j < spawnCount; j++)
				{

					int id = findFreeParticle();

					if (id == -1)
						return;
					Particle& particle = particles[id];
					particle.lifeTime = sf::seconds(lifeSeconds + getRandomValue11() * r_lifeTime);
					particle.totalLifeTime = particle.lifeTime;
					sf::Color startRandom;
					if (r_useSecondary)
						startRandom = sf::Color(startColor.r + getRandomValue01() * (r_secondaryStart.r - startColor.r),
							startColor.g + getRandomValue01() * (r_secondaryStart.g - startColor.g),
							startColor.b + getRandomValue01() * (r_secondaryStart.b - startColor.b));
					else
						startRandom = startColor;

					sf::Color endRandom;
					if (r_useSecondary)
						endRandom = sf::Color(endColor.r + getRandomValue01() * (r_secondaryEnd.r - endColor.r),
							endColor.g + getRandomValue01() * (r_secondaryEnd.g - endColor.g),
							endColor.b + getRandomValue01() * (r_secondaryEnd.b - endColor.b));
					else
						endRandom = endColor;
					particle.currentColor = startRandom;
					particle.startColor = startRandom;
					particle.endColor = endRandom;
					particle.force = startForce;
					particle.inUse = 1;
					particle.startSize = startSize + getRandomValue11() * r_startSize;
					particle.endSize = endSize + getRandomValue11() * r_endSize;
					particle.startSpeed = startSpeed + getRandomValue11() * r_startSpeed;
					particle.endSpeed = endSpeed + getRandomValue11() * r_endSpeed;

					float randomDirection = getRandomValue01() * fireAngle + fireRotation;
					particle.direction = sf::Vector2f(cos(randomDirection * 0.0174533), sin(randomDirection * 0.0174533));
					particle.setRotation(randomStartRotation ? getRandomValue01() * 360 : 0);
					float random1 = getRandomValueTrig();
					float random2 = getRandomValue01();
					sf::Vector2f randomPos = startPos + sf::Vector2f(
						cos(random1) * random2 * spawnRadius,
						sin(random1) * random2 * spawnRadius);
					particle.offset = randomPos;
					particle.setPosition(randomPos);
					particle.Update();
					totalParticles++;

					createEvent(ParticleSystemEvent::Type::OnCreate, particle);
				}
			}
			refreshPointer();
			return;

		}
		void Update(const sf::Time& deltaTime)
		{
			updated2->lock();
			sf::Clock clock;
			if (createOnUpdate)
			{
				cntDown -= deltaTime;

				while (cntDown.asSeconds() <= 0)
				{
					Create();
					cntDown += sf::seconds(spawnSeconds);
					if (!keepUpWithFrameRate)
						break;
				}

			}
			createFor -= deltaTime;
			if (createFor.asSeconds() > 0)
				Create();
			for (int i = 0; i < particles.size(); i++)
			{
				if (!particles[i].inUse)
					continue;
				//update
				Particle& p = particles[i];
				p.lifeTime -= deltaTime;
			}
			int cnt = 0;
			if (drawNewestOnTop)
			{
				cnt = 0;
				for (int i = 0; i < particles.size(); i++)
				{
					if (!particles[i].inUse)
						continue;
					Particle& p = particles[i];
					if (p.lifeTime.asSeconds() <= 0)
					{
						cnt++;
						createEvent(ParticleSystemEvent::Type::OnDeath, p);
						p.currentColor = sf::Color::Transparent;
						p.Update();
					}
					else
						break;
				}
				if (cnt > 0)
				{
					//kill particle
					for (int j = 0; j < verticesPerParticle(); j++)
						arr.erase(arr.begin(), arr.begin() + cnt);
					particles.erase(particles.begin(), particles.begin() + cnt);
					totalParticles -= cnt;
					if (totalParticles < 0)
						totalParticles = 0;
					refreshPointer();
				}
			}
			else
			{
				for (int i = 0; i < particles.size(); i++)
				{
					if (!particles[i].inUse)
						continue;
					Particle& p = particles[i];
					if (p.lifeTime.asSeconds() <= 0)
					{
						createEvent(ParticleSystemEvent::Type::OnDeath, p);
						p.currentColor = sf::Color::Transparent;
						p.Update();
						totalParticles--;
						if (totalParticles < 0)
							totalParticles = 0;
						p.inUse = 0;
					}
				}
			}
			//update
			for (int i = 0; i < particles.size(); i++)
			{
				if (!particles[i].inUse)
					continue;
				Particle& p = particles[i];
				p.force += constantForce * deltaTime.asSeconds();
				p.offset += (p.direction + p.force) * deltaTime.asSeconds() * getParam(p.totalLifeTime, p.lifeTime, p.startSpeed, p.endSpeed);
				p.setPosition(p.offset);

				if (!randomStartRotation)
					p.setRotation(getParam(p.totalLifeTime, p.lifeTime, startRotation, endRotation));
				p.setScale(sf::Vector2f(1, 1) * getParam(p.totalLifeTime, p.lifeTime, p.startSize, p.endSize));

				float zeroToOne = std::min((p.totalLifeTime.asSeconds() - p.lifeTime.asSeconds()) / p.totalLifeTime.asSeconds(), 1.f);
				float fadingT = p.currentColor.a;
				if (zeroToOne > fading && fading < 1)
				{
					zeroToOne -= fading;
					zeroToOne /= (1 - fading);
					fadingT = 255 - 255 * zeroToOne;
				}
				p.currentColor = sf::Color(getParam(p.totalLifeTime, p.lifeTime, p.startColor.r, p.endColor.r),
					getParam(p.totalLifeTime, p.lifeTime, p.startColor.g, p.endColor.g),
					getParam(p.totalLifeTime, p.lifeTime, p.startColor.b, p.endColor.b), fadingT);
				createEvent(ParticleSystemEvent::Type::OnUpdate, p);
			}
			for (int i = 0; i < particles.size(); i++)
			{
				if (!particles[i].inUse)
					continue;
				Particle& p = particles[i];
				p.Update();
			}
			updated2->unlock();
			{
				sf::Lock lock(*updated);
				if (arr.size() > 0)
				{
					if (buff.getVertexCount() != arr.size())
						buff.create(arr.size());
					buff.update(&arr[0]);
				}
			}
			updateTime = clock.getElapsedTime();
		}
	private:
		virtual void draw(sf::RenderTarget& target, sf::RenderStates states) const
		{
			states.transform *= getTransform();
			states.texture = texture;
			if (arr.size() != 0)
			{
				if (sf::VertexBuffer::isAvailable())
				{
					sf::Lock lock(*updated);
					target.draw(buff, states);
				}
				else
				{
					sf::Lock lock2(*updated2);
					if (type == ParticleType::QuadsTriangles)
						target.draw(&arr[0], arr.size(), sf::PrimitiveType::Triangles, states);
					else
						target.draw(&arr[0], arr.size(), static_cast<sf::PrimitiveType>(type), states);
				}

			}
		}
	};
	enum class ObjectType
	{
		None,
		Tilemap,
		Sprite,
		Texture,
		Shader_F,
		Shader_V,
		ParticleSystem,
	};
	class Level : public sf::Drawable, public sf::Transformable
	{
	public:
		struct Uniforms
		{
			enum class DataTypes
			{
				Bool,
				Float,
				Int,
				Vec2,
				Vec3,
				Vec4,
				IVec2,
				IVec3,
				IVec4,
			};
			std::vector<DataTypes> types;
			std::vector<std::string> uniformNames;
			unsigned int byteSize;
			static int getTypeSize(DataTypes type)
			{
				switch (type)
				{
				case Uniforms::DataTypes::Bool:
					return 1;
					break;
				case Uniforms::DataTypes::Float:
					return 4;
					break;
				case Uniforms::DataTypes::Int:
					return 4;
					break;
				case Uniforms::DataTypes::Vec2:
					return 8;
					break;
				case Uniforms::DataTypes::Vec3:
					return 12;
					break;
				case Uniforms::DataTypes::Vec4:
					return 16;
					break;
				case Uniforms::DataTypes::IVec2:
					return 8;
					break;
				case Uniforms::DataTypes::IVec3:
					return 12;
					break;
				case Uniforms::DataTypes::IVec4:
					return 16;
					break;
				default:
					break;
				}
				return 0;
			}
			unsigned int getIndex(const std::string& name)
			{
				for (int i = 0; i < uniformNames.size(); i++)
				{
					if (name == uniformNames[i])
					{
						return i;
					}
				}
				return 0;
			}
			DataTypes getUniformType(unsigned int index) const
			{
				return types[index];
			}
			int getOffsetToIndex(int index) const
			{
				if (index < 0)
					return -1;
				int offset = 0;
				for (int i = 0; i < index; i++)
				{
					offset += getTypeSize(types[i]);
				}
				return offset;
			}
			static unsigned char* getUniform(unsigned int offset, std::vector<unsigned char>* data)
			{
				return &((*data)[offset]);
			}
			int getIndexFromName(const std::string& name)
			{
				for (int i = 0; i < uniformNames.size(); i++)
				{
					if (uniformNames[i] == name)
						return i;
				}
				return -1;
			}
			void setUniform(unsigned int index, void* uniformPointer, sf::Shader& shader)
			{
				switch (getUniformType(index))
				{
				case Uniforms::DataTypes::Bool:
					shader.setUniform(uniformNames[index], *((bool*)uniformPointer));
					break;
				case Uniforms::DataTypes::Float:
					shader.setUniform(uniformNames[index], *((float*)uniformPointer));
					break;
				case Uniforms::DataTypes::Int:
					shader.setUniform(uniformNames[index], *((int*)uniformPointer));
					break;
				case Uniforms::DataTypes::Vec2:
					shader.setUniform(uniformNames[index], *((sf::Glsl::Vec2*)uniformPointer));
					break;
				case Uniforms::DataTypes::Vec3:
					shader.setUniform(uniformNames[index], *((sf::Glsl::Vec3*)uniformPointer));
					break;
				case Uniforms::DataTypes::Vec4:
					shader.setUniform(uniformNames[index], *((sf::Glsl::Vec4*)uniformPointer));
					break;
				case Uniforms::DataTypes::IVec2:
					shader.setUniform(uniformNames[index], *((sf::Glsl::Ivec2*)uniformPointer));
					break;
				case Uniforms::DataTypes::IVec3:
					shader.setUniform(uniformNames[index], *((sf::Glsl::Ivec3*)uniformPointer));
					break;
				case Uniforms::DataTypes::IVec4:
					shader.setUniform(uniformNames[index], *((sf::Glsl::Ivec4*)uniformPointer));
					break;
				default:
					break;
				}
			}
		};
		struct RenderStateObject
		{
		private:
			friend class Level;
			std::unique_ptr<sf::Texture> texture;
			std::string locationData;
			std::unique_ptr<Uniforms> uniforms;
			ObjectType type;


		public:
			RenderStateObject()
			{
				type = ObjectType::None;
			}
			~RenderStateObject()
			{

			}
			bool UseTexture()
			{
				if (texture || uniforms)
					return 0;
				texture = std::make_unique<sf::Texture>();
				return 1;
			}
			bool UseUniforms()
			{
				if (texture || uniforms)
					return 0;
				uniforms = std::make_unique<Uniforms>();
				return 1;
			}
			const ObjectType& getObjectType() const
			{
				return type;
			}
			const sf::Texture& getTexture() const
			{
				return *texture.get();
			}
			const std::string& getTextureLocation() const
			{
				return locationData;
			}
		};
		struct Object : public Drawable
		{
		private:
			friend class Level;
			std::unique_ptr<sf::Sprite> sprite;
			std::unique_ptr<TileMap> tilemap;
			std::unique_ptr<ParticleSystem> particleSystem;
			sf::Texture* usingTexture = nullptr;
			mutable sf::Shader shader;
			mutable std::vector<unsigned char> uniformFData;
			mutable std::vector<unsigned char> uniformVData;
			ObjectType type;
			sf::Int8 drawingZOrder;
			bool visible = 1;
			sf::BlendMode blendMode;
			RenderStateObject* useVShader;
			RenderStateObject* useFShader;
			sf::Color color;

			bool UseSprite()
			{
				if (tilemap || sprite || particleSystem)
					return 0;
				sprite = std::make_unique<sf::Sprite>();
				return 1;
			}
			bool UseTileMap()
			{
				if (tilemap || sprite || particleSystem)
					return 0;
				tilemap = std::make_unique<TileMap>();
				return 1;
			}
			bool UseParticleSystem()
			{
				if (tilemap || sprite || particleSystem)
					return 0;
				particleSystem = std::make_unique<ParticleSystem>();
				return 1;
			}
			void* getPointer() const
			{
				if (tilemap)
					return tilemap.get();
				if (sprite)
					return sprite.get();
				if (particleSystem)
					return particleSystem.get();
			}
		public:
			Object()
				: drawingZOrder(0)
			{
				sprite = nullptr;
				tilemap = nullptr;
				particleSystem = nullptr;
				useVShader = nullptr;
				useFShader = nullptr;
				blendMode = sf::BlendAlpha;
			}
			~Object()
			{
			}
			const sf::Int8& getDrawingZLayer() const
			{
				return drawingZOrder;
			}
			const bool& isVisible() const
			{
				return visible;
			}
			const ObjectType& getObjectType() const
			{
				return type;
			}
			const sf::BlendMode& getBlendMode() const
			{
				return blendMode;
			}
			Transformable& getTransformable() const
			{
				if (tilemap)
					return (*tilemap);
				if (sprite)
					return (*sprite);
				if (particleSystem)
					return (*particleSystem);
				sf::err() << "Object does not have a transformable!" << std::endl;
				return (*sprite);
			}
			sf::Sprite& getSprite() const
			{
				if (!sprite)
					sf::err() << "Object not a sprite!" << std::endl;
				return (*sprite);
			}
			TileMap& getTileMap() const
			{
				if (!tilemap)
					sf::err() << "Object not a tilemap!" << std::endl;
				return (*tilemap);
			}
			ParticleSystem& getParticleSystem() const
			{
				if (!particleSystem)
					sf::err() << "Object not a particle system!" << std::endl;
				return (*particleSystem);
			}
			const sf::Shader& getShader() const
			{
				return shader;
			}
			void setDrawingLayer(sf::Int8 drawingLayer)
			{
				drawingZOrder = drawingLayer;
			}
			void setVisible(bool visible)
			{
				this->visible = visible;
			}
			void setBlendMode(const sf::BlendMode& blendMode)
			{
				this->blendMode = blendMode;
			}
			template<typename T>
			void setUniform(const std::string& name, T value)
			{
				shader.setUniform(name, value);
			}

		private:
			virtual void draw(sf::RenderTarget& target, sf::RenderStates states) const
			{
				states.blendMode = blendMode;
				if (sprite)
					target.draw(*sprite, states);
				if (tilemap)
					target.draw(*tilemap, states);
				if (particleSystem)
					target.draw(*particleSystem, states);
			}
		};
	protected:
		static void AnalyzeShader(RenderStateObject& targetObject)
		{
			if (targetObject.type != ObjectType::Shader_F && targetObject.type != ObjectType::Shader_V)
				return;
			sf::FileInputStream load1;
			if (load1.open(targetObject.locationData))
			{
				auto stringToType = [](const std::string& str)
				{
					if (str == "bool")
						return Uniforms::DataTypes::Bool;
					if (str == "int")
						return Uniforms::DataTypes::Int;
					if (str == "float")
						return Uniforms::DataTypes::Float;
					if (str == "ivec2")
						return Uniforms::DataTypes::IVec2;
					if (str == "ivec3")
						return Uniforms::DataTypes::IVec3;
					if (str == "ivec4")
						return Uniforms::DataTypes::IVec4;
					if (str == "vec2")
						return Uniforms::DataTypes::Vec2;
					if (str == "vec3")
						return Uniforms::DataTypes::Vec3;
					if (str == "vec4")
						return Uniforms::DataTypes::Vec4;
				};
				int byteSize = 0;
				targetObject.uniforms->types.clear();
				targetObject.uniforms->uniformNames.clear();
				while (1)
				{
					std::string file;
					LoadLine(load1, file);
					if (file.empty())
						break;
					if (file == "uniform")
					{
						LoadLine(load1, file);
						if (file != "sampler2D")
						{
							Uniforms::DataTypes type = stringToType(file);
							byteSize += Uniforms::getTypeSize(type);
							targetObject.uniforms->types.push_back(type);
							LoadLine(load1, file);
							if (file[file.size() - 1] == ';')
							{
								file.pop_back();
							}
							targetObject.uniforms->uniformNames.push_back(file);
						}
					}
				}
				targetObject.uniforms->byteSize = byteSize;
			}
		}
		template<typename T>
		static void Save(std::string& s, T var)
		{
			for (int i = 0; i < sizeof(T); i++)
				s += (reinterpret_cast<char*>(&var))[i];
		}
		template<typename T>
		static void Save(std::ofstream& s, T var)
		{
			for (int i = 0; i < sizeof(T); i++)
				s << (reinterpret_cast<char*>(&var))[i];
		}
		static void SaveStr(std::ofstream& s, const std::string& var)
		{
			sf::Uint32 size = var.size();
			for (int i = 0; i < sizeof(sf::Uint32); i++)
				s << (reinterpret_cast<char*>(&size))[i];
			s << var;
		}
		template<typename T>
		static void Load(sf::InputStream& input, T& t)
		{
			input.read(&t, sizeof(T));
		}
		static void LoadStr(sf::InputStream& input, std::string& data)
		{
			sf::Uint32 size;
			input.read(&size, sizeof(sf::Uint32));
			data.resize(size);
			input.read(&data[0], size);
		}
		static void LoadLine(sf::InputStream& input, std::string& data)
		{
			char x;
			data.clear();
			while (1)
			{
				if (input.read(&x, 1) == -1)
					break;
				if (x == '\r')
					continue;
				if (x == ' ' || x == '\n')
					break;
				data.push_back(x);
			}
		}
		std::map<std::string, Object> objects;
		std::map<std::string, RenderStateObject> texture_objects;

	public:
		Level()
		{
		}
		~Level()
		{

		}
		
		const unsigned int getLevelObjectCount() const
		{
			return objects.size();
		}
		Object& getLevelObject(const std::string& name)
		{
			return objects.at(name);
		}
		const std::map<std::string, Object>& getLevelMap() const
		{
			return objects;
		}
		const unsigned int getRenderStateObjectCount() const
		{
			return texture_objects.size();
		}
		RenderStateObject& getRenderStateObject(const std::string& name)
		{
			return texture_objects.at(name);
		}
		const std::map<std::string, RenderStateObject>& getRenderStateMap() const
		{
			return texture_objects;
		}
		void loadFromFile(const std::string& fileName, const std::string& assetFolder)
		{
			sf::BlendMode blendModes[4] = { sf::BlendMode(sf::BlendMode::One, sf::BlendMode::Zero),
			sf::BlendMode(sf::BlendMode::SrcAlpha, sf::BlendMode::One, sf::BlendMode::Add, sf::BlendMode::One, sf::BlendMode::One, sf::BlendMode::Add),
			sf::BlendMode(sf::BlendMode::SrcAlpha, sf::BlendMode::OneMinusSrcAlpha, sf::BlendMode::Add, sf::BlendMode::One, sf::BlendMode::OneMinusSrcAlpha, sf::BlendMode::Add),
			sf::BlendMode(sf::BlendMode::DstColor, sf::BlendMode::Zero) };
			sf::FileInputStream load1;
			if (load1.open(fileName))
			{
				std::string line;
				LoadStr(load1, line);
				if (line != ZLE_VERSION)
					return;
				int counter;
				int temp;
				float tempF;
				bool tempB;
				Load(load1, counter);
				std::vector<sf::Texture*> textureP;
				std::vector<RenderStateObject*> shaderP;
				textureP.resize(counter);
				shaderP.resize(counter);
				for (int i = 0; i < counter; i++)
				{
					ObjectType otmp;
					Load(load1, otmp);

					LoadStr(load1, line);
					if (otmp != ObjectType::Texture && otmp != ObjectType::Shader_V && otmp != ObjectType::Shader_F)
					{
						Load(load1, tempB);
						Load(load1, tempB);
						continue;
					}
					std::string line2 = line;
					for (int i = 0; i < UINT16_MAX; i++)
					{
						if (line2[0] == '\\' || line2[0] == '/')
						{
							line2.erase(0, 1);
							break;
						}
						else
							line2.erase(0, 1);

					}
					std::string line3 = line2;
					for (int i = 0; i < UINT16_MAX; i++)
					{
						if (line3[0] == '\\' || line3[0] == '/')
						{
							line3.erase(0, 1);
							break;
						}
						else
							line3.erase(0, 1);

					}
					texture_objects[line2].type = otmp;
					if (texture_objects[line2].type == ObjectType::Texture)
						texture_objects[line2].UseTexture();
					texture_objects[line2].locationData = assetFolder + line3;
					if (texture_objects[line2].type == ObjectType::Texture)
						texture_objects[line2].texture->loadFromFile(assetFolder + line3);
					Load(load1, tempB);
					if (texture_objects[line2].type == ObjectType::Texture)
						texture_objects[line2].texture->setSmooth(tempB);
					Load(load1, tempB);
					if (texture_objects[line2].type == ObjectType::Texture)
						texture_objects[line2].texture->setRepeated(tempB);
					if (texture_objects[line2].type != ObjectType::Texture)
					{
						texture_objects[line2].UseUniforms();
						AnalyzeShader(texture_objects[line2]);
					}
					textureP[i] = texture_objects[line2].texture.get();
					shaderP[i] = &texture_objects[line2];

				}
				Load(load1, counter);
				for (int i = 0; i < counter; i++)
				{
					LoadStr(load1, line);
					Object& o = objects[line];
					Load(load1, o.color);
					Load(load1, o.type);
					switch (o.type)
					{
					case ObjectType::Sprite:
						o.UseSprite();
						break;
					case ObjectType::Tilemap:
						o.UseTileMap();
						break;
					case ObjectType::ParticleSystem:
						o.UseParticleSystem();
						break;
					}
					Load(load1, temp);
					if (temp != -1)
						o.usingTexture = textureP[temp];

					Load(load1, temp);
					if (temp != -1)
					{
						o.useFShader = shaderP[temp];
						size_t size;
						Load(load1, size);
						char temp2;
						o.uniformFData.resize(size);
						for (int i = 0; i < size; i++)
						{
							Load(load1, temp2);
							o.uniformFData[i] = temp2;
						}
					}

					Load(load1, temp);
					if (temp != -1)
					{
						o.useVShader = shaderP[temp];

						size_t size;
						Load(load1, size);
						char temp2;
						o.uniformVData.resize(size);
						for (int i = 0; i < size; i++)
						{
							Load(load1, temp2);
							o.uniformVData[i] = temp2;
						}
					}
					if (o.useVShader && o.useFShader)
						o.shader.loadFromFile(o.useVShader->locationData, o.useFShader->locationData);
					else if (o.useFShader && !o.useVShader)
						o.shader.loadFromFile(o.useFShader->locationData, sf::Shader::Fragment);
					else if (!o.useFShader && o.useVShader)
						o.shader.loadFromFile(o.useVShader->locationData, sf::Shader::Vertex);

					Load(load1, temp);
					o.visible = temp;
					Load(load1, temp);
					o.drawingZOrder = temp;
					Load(load1, temp);
					o.blendMode = blendModes[temp];

					sf::Vector2f tempV;
					sf::Transformable& t = o.getTransformable();
					Load(load1, tempV);
					t.setPosition(tempV);

					Load(load1, tempF);
					t.setRotation(tempF);

					Load(load1, tempV);
					t.setScale(tempV);

					Load(load1, tempV);
					t.setOrigin(tempV);

					switch (o.type)
					{
					case ObjectType::Sprite:
					{
						sf::Sprite* s = o.sprite.get();
						s->setColor(o.color);

						sf::IntRect tempI;
						Load(load1, tempI);
						s->setTextureRect(tempI);
						if (o.usingTexture)
							s->setTexture(*o.usingTexture);
					}
					break;
					case ObjectType::Tilemap:
					{
						TileMap* s = o.tilemap.get();
						unsigned int tempU;
						sf::Vector2u tempVU;

						Load(load1, tempU);
						s->setTextureGap(tempU);

						Load(load1, tempVU);
						s->setTextureTileSize(tempVU);
						Load(load1, tempVU);
						s->setTileCount(tempVU);
						Load(load1, tempVU);
						s->setTileSize(tempVU);

						s->setColor(o.color);
						sf::Uint16 tileID;
						for (int i = 0; i < s->getTileCount().x; i++)
							for (int j = 0; j < s->getTileCount().y; j++)
							{
								Load(load1, tileID);
								s->setTileID(sf::Vector2u(i, j), tileID);
							}
						if (o.usingTexture)
							s->setTexture(o.usingTexture);
					}
					break;
					case ObjectType::ParticleSystem:
					{
						o.particleSystem->loadFromStream(load1);
						if (o.usingTexture)
							o.particleSystem->setTexture(*o.usingTexture);
					}
					break;
					}

				}
			}
		}
		void clear()
		{
			objects.clear();
			texture_objects.clear();
		}

	private:
		virtual void draw(sf::RenderTarget& target, sf::RenderStates states) const
		{
			states.transform *= getTransform();
			for (int i = -100; i <= 100; i++)
			{
				for (auto& n : objects)
				{
					if (!n.second.visible)
						continue;
					if (n.second.drawingZOrder == i)
					{
						if (n.second.useFShader || n.second.useVShader)
							states.shader = &n.second.shader;
						if (!n.second.useFShader && !n.second.useVShader)
							states.shader = nullptr;
						target.draw(n.second, states);
					}
				}
			}
		}
	};

}
namespace squ
{
	static const std::string SQU_VERSION = "SQU002";
	static const std::string SQV_VERSION = "SQV002";
	class Frame : public sf::Drawable
	{
		friend class FrameAnimation;
		sf::VertexArray arr;
		sf::VertexBuffer buff;
		sf::Vector2u size;
		const sf::Texture* texture;
		enum Settings
		{
			None = 0,
			ConstRotation = 1 << 0,
			ConstSize = 1 << 1,
		};
		static void readObject(sf::InputStream& stream, sf::Vector2i& position, sf::Uint16& rotation, sf::Uint16& scale, sf::Uint32 settings, sf::Uint16 conRot, sf::Uint16 conSize)
		{
			if (settings == None)
			{
				sf::Uint8 byte[6];
				for (int i = 0; i < 6; i++)
					stream.read(&byte[i], sizeof(byte[i]));
				scale = byte[0] | ((byte[1] & 0b111) << 8);
				rotation = ((byte[1] & 0b11111000) >> 3) | ((byte[2] & 0b1111) << 5);
				position.x = (((byte[2] & 0b11110000) >> 4) | (byte[3] << 4) | ((byte[4] & 0b11) << 12)) - 8192;
				position.y = (((byte[4] & 0b11111100) >> 2) | (byte[5] << 6)) - 8192;
			}
			else if (settings & ConstRotation && settings & ConstSize)
			{
				sf::Uint8 byte[4];
				for (int i = 0; i < 4; i++)
					stream.read(&byte[i], sizeof(byte[i]));
				scale = conSize;
				rotation = conRot;
				position.x = ((byte[1] << 8) | byte[0]) - 8192;
				position.y = ((byte[3] << 8) | byte[2]) - 8192;
			}
			else if (settings == ConstRotation)
			{
				sf::Uint8 byte[5];
				for (int i = 0; i < 5; i++)
					stream.read(&byte[i], sizeof(byte[i]));
				scale = byte[0] | ((byte[1] & 0b1111) << 8);
				rotation = conRot;
				position.x = (((byte[1] & 0b11110000) >> 4) | (byte[2] << 4) | ((byte[3] & 0b11) << 12)) - 8192;
				position.y = (((byte[3] & 0b11111100) >> 2) | (byte[4] << 6)) - 8192;
			}
			else if (settings == ConstSize)
			{
				sf::Uint8 byte[5];
				for (int i = 0; i < 5; i++)
					stream.read(&byte[i], sizeof(byte[i]));
				scale = conSize;
				rotation = byte[0] | ((byte[1] & 0b1111) << 8);
				position.x = (((byte[1] & 0b11110000) >> 4) | (byte[2] << 4) | ((byte[3] & 0b11) << 12)) - 8192;
				position.y = (((byte[3] & 0b11111100) >> 2) | (byte[4] << 6)) - 8192;
			}
		}
	public:
		Frame()
			: size(0, 0), texture(nullptr)
		{

		}
		const sf::Vector2u& getSize() const { return size; }
		const sf::Texture* getTexture() const { return texture; }
		void setTexture(const sf::Texture* texture)
		{
			if (texture != this->texture && texture)
			{
				for (int i = 0; i < arr.getVertexCount() / 6; i++)
				{
					arr[i * 6 + 0].texCoords = sf::Vector2f(0, 0);
					arr[i * 6 + 1].texCoords = sf::Vector2f(texture->getSize().x, 0);
					arr[i * 6 + 2].texCoords = sf::Vector2f(texture->getSize().x, texture->getSize().y);
					arr[i * 6 + 3].texCoords = sf::Vector2f(0, texture->getSize().y);

					arr[i * 6 + 4].texCoords = arr[i * 6 + 0].texCoords;
					arr[i * 6 + 5].texCoords = arr[i * 6 + 2].texCoords;
				}
			}
			this->texture = texture;
		}
		bool loadFromStream(sf::InputStream& stream)
		{
			sf::Uint32 settings = 0;
			sf::Uint16 conRot = 0;
			sf::Uint16 conSize = 0;
			sf::Int16 sizeX;
			sf::Int16 sizeY;
			sf::Uint32 count;
			std::string ver = "123456";
			stream.read(&ver[0], 6);
			if (ver != SQU_VERSION)
				return false;
			stream.read(&sizeX, sizeof(sizeX));
			stream.read(&sizeY, sizeof(sizeY));
			size.x = sizeX;
			size.y = sizeY;
			stream.read(&count, sizeof(count));
			stream.read(&settings, sizeof(settings));
			if (settings & ConstRotation)
				stream.read(&conRot, sizeof(conRot));
			if (settings & ConstSize)
				stream.read(&conSize, sizeof(conSize));

			arr.setPrimitiveType(sf::Triangles);
			arr.resize(count * 6);

			sf::Vector2i position;
			sf::Uint16 rotation = 0;
			sf::Uint16 scale = 0;
			sf::Color color;
			sf::Transformable tran;

			for (int i = 0; i < count; i++)
			{
				readObject(stream, position, rotation, scale, settings, conRot, conSize);

				stream.read(&color.r, sizeof(color.r));
				stream.read(&color.g, sizeof(color.g));
				stream.read(&color.b, sizeof(color.b));

				tran.setPosition(position.x, position.y);
				tran.setRotation(360 - rotation / 512.f * 360.f);
				tran.setScale(scale, scale);

				arr[i * 6 + 0].position = tran.getTransform().transformPoint(sf::Vector2f(-0.5, -0.5));
				arr[i * 6 + 0].color = color;
				arr[i * 6 + 1].position = tran.getTransform().transformPoint(sf::Vector2f(0.5, -0.5));
				arr[i * 6 + 1].color = color;
				arr[i * 6 + 2].position = tran.getTransform().transformPoint(sf::Vector2f(0.5, 0.5));
				arr[i * 6 + 2].color = color;
				arr[i * 6 + 3].position = tran.getTransform().transformPoint(sf::Vector2f(-0.5, 0.5));
				arr[i * 6 + 3].color = color;

				if (texture)
				{
					arr[i * 6 + 0].texCoords = sf::Vector2f(0, 0);
					arr[i * 6 + 1].texCoords = sf::Vector2f(texture->getSize().x, 0);
					arr[i * 6 + 2].texCoords = sf::Vector2f(texture->getSize().x, texture->getSize().y);
					arr[i * 6 + 3].texCoords = sf::Vector2f(0, texture->getSize().y);
				}

				arr[i * 6 + 4] = arr[i * 6 + 0];
				arr[i * 6 + 5] = arr[i * 6 + 2];
			}
			if (sf::VertexBuffer::isAvailable() && arr.getVertexCount() > 0)
			{
				buff.setUsage(sf::VertexBuffer::Static);
				buff.setPrimitiveType(arr.getPrimitiveType());
				buff.create(arr.getVertexCount());
				buff.update(&arr[0]);
			}
			return true;
		}
		bool loadFromFile(const std::string& fileName)
		{
			sf::FileInputStream file;
			if (file.open(fileName))
				return loadFromStream(file);
			return false;
		}
	private:
		virtual void draw(sf::RenderTarget& target, sf::RenderStates states) const
		{
			if (texture)
				states.texture = texture;
			if (sf::VertexBuffer::isAvailable())
				target.draw(buff, states);
			else
				target.draw(arr, states);
		}
	};
	class FrameAnimation
	{
		friend class FrameAnimator;
		std::vector<sf::VertexArray> arr;
		std::vector<sf::VertexBuffer> buff;
		sf::Vector2u size;
		sf::Uint16 frameCount;
		const sf::Texture* texture;
	public:
		FrameAnimation()
			: size(0, 0), texture(nullptr), frameCount(0)
		{

		}
		const sf::Vector2u& getSize() const { return size; }
		const sf::Texture* getTexture() const { return texture; }
		const sf::Uint16& getFrameCount() const { return frameCount; }
		void setTexture(const sf::Texture* texture)
		{
			if (texture != this->texture && texture)
			{
				for (int j = 0; j < arr.size(); j++)
					for (int i = 0; i < arr[j].getVertexCount() / 6; i++)
					{
						arr[j][i * 6 + 0].texCoords = sf::Vector2f(0, 0);
						arr[j][i * 6 + 1].texCoords = sf::Vector2f(texture->getSize().x, 0);
						arr[j][i * 6 + 2].texCoords = sf::Vector2f(texture->getSize().x, texture->getSize().y);
						arr[j][i * 6 + 3].texCoords = sf::Vector2f(0, texture->getSize().y);

						arr[j][i * 6 + 4].texCoords = arr[j][i * 6 + 0].texCoords;
						arr[j][i * 6 + 5].texCoords = arr[j][i * 6 + 2].texCoords;
					}
			}
			this->texture = texture;
		}
		bool loadFromStream(sf::InputStream& stream)
		{
			sf::Uint32 settings = 0;
			sf::Uint16 conRot = 0;
			sf::Uint16 conSize = 0;
			sf::Uint16 sizeX;
			sf::Uint16 sizeY;
			sf::Uint16 count;
			std::string ver = "123456";
			stream.read(&ver[0], 6);
			if (ver != SQV_VERSION)
				return false;
			stream.read(&sizeX, sizeof(sizeX));
			stream.read(&sizeY, sizeof(sizeY));
			stream.read(&count, sizeof(count));
			size.x = sizeX;
			size.y = sizeY;

			arr.resize(count);
			if (sf::VertexBuffer::isAvailable())
				buff.resize(count);
			frameCount = count;

			sf::Vector2i position;
			sf::Uint16 rotation = 0;
			sf::Uint16 scale = 0;
			sf::Color color;
			sf::Transformable tran;

			for (int i = 0; i < count; i++)
			{
				sf::Uint32 objectCount = 0;
				stream.read(&objectCount, sizeof(objectCount));
				arr[i].setPrimitiveType(sf::Triangles);
				arr[i].resize(objectCount * 6);
				stream.read(&settings, sizeof(settings));
				if (settings & Frame::ConstRotation)
					stream.read(&conRot, sizeof(conRot));
				if (settings & Frame::ConstSize)
					stream.read(&conSize, sizeof(conSize));

				for (int j = 0; j < objectCount; j++)
				{
					Frame::readObject(stream, position, rotation, scale, settings, conRot, conSize);

					stream.read(&color.r, sizeof(color.r));
					stream.read(&color.g, sizeof(color.g));
					stream.read(&color.b, sizeof(color.b));

					tran.setPosition(position.x, position.y);
					tran.setRotation(360 - rotation / 512.f * 360.f);
					tran.setScale(scale, scale);

					arr[i][j * 6 + 0].position = tran.getTransform().transformPoint(sf::Vector2f(-0.5, -0.5));
					arr[i][j * 6 + 0].color = color;
					arr[i][j * 6 + 1].position = tran.getTransform().transformPoint(sf::Vector2f(0.5, -0.5));
					arr[i][j * 6 + 1].color = color;
					arr[i][j * 6 + 2].position = tran.getTransform().transformPoint(sf::Vector2f(0.5, 0.5));
					arr[i][j * 6 + 2].color = color;
					arr[i][j * 6 + 3].position = tran.getTransform().transformPoint(sf::Vector2f(-0.5, 0.5));
					arr[i][j * 6 + 3].color = color;

					if (texture)
					{
						arr[i][j * 6 + 0].texCoords = sf::Vector2f(0, 0);
						arr[i][j * 6 + 1].texCoords = sf::Vector2f(texture->getSize().x, 0);
						arr[i][j * 6 + 2].texCoords = sf::Vector2f(texture->getSize().x, texture->getSize().y);
						arr[i][j * 6 + 3].texCoords = sf::Vector2f(0, texture->getSize().y);
					}

					arr[i][j * 6 + 4] = arr[i][j * 6 + 0];
					arr[i][j * 6 + 5] = arr[i][j * 6 + 2];
				}
				if (sf::VertexBuffer::isAvailable() && arr[i].getVertexCount() > 0)
				{
					buff[i].setUsage(sf::VertexBuffer::Static);
					buff[i].setPrimitiveType(arr[i].getPrimitiveType());
					buff[i].create(arr[i].getVertexCount());
					buff[i].update(&arr[i][0]);
				}
			}
			return true;
		}
		bool loadFromFile(const std::string& fileName)
		{
			sf::FileInputStream file;
			if (file.open(fileName))
				return loadFromStream(file);
			return false;
		}

	};
	class FrameAnimator : public sf::Drawable, public sf::Transformable
	{
	public:
		enum Status
		{
			Stopped,
			Paused,
			Playing,
		};
	private:
		Status status;
		const FrameAnimation* anim;
		sf::Uint16 atFrame;
		sf::Uint8 fps;
		sf::Time offset;
		bool loop;
	public:
		FrameAnimator()
			: anim(nullptr), atFrame(0), fps(10), offset(sf::Time::Zero), loop(false), status(Stopped)
		{

		}
		const FrameAnimation* getAnimation() const { return anim; }
		const Status& getStatus() const { return status; }
		const sf::Uint16& getPlayingFrame() const { return atFrame; }
		const sf::Time& getPlayingOffset() const { return offset; }
		const sf::Uint8& getFramesPerSecond() const { return fps; }
		const bool& getLoop() const { return loop; }

		void setLoop(bool state)
		{
			loop = state;
		}
		void play()
		{
			if (status == Stopped)
			{
				atFrame = 0;
				offset = sf::Time::Zero;
			}
			status = Playing;
		}
		void pause()
		{
			status = Paused;
		}
		void stop()
		{
			atFrame = 0;
			offset = sf::Time::Zero;
			status = Stopped;
		}
		void setFramesPerSecond(sf::Uint8 fps)
		{
			this->fps = fps;
		}
		void setAnimation(const FrameAnimation* animation)
		{
			this->anim = animation;
		}
		void update(const sf::Time& deltaTime)
		{
			if (status != Playing)
				return;
			if (!anim)
				return;
			offset += deltaTime;
			atFrame = offset.asSeconds() * fps;
			if (atFrame >= anim->getFrameCount())
			{
				if (!loop)
					status = Stopped;
				else
				{
					atFrame = 0;
					offset = sf::Time::Zero;
				}
			}
		}
	private:
		virtual void draw(sf::RenderTarget& target, sf::RenderStates states) const
		{
			if (!anim)
				return;
			if (atFrame >= anim->getFrameCount())
				return;
			if (anim->texture)
				states.texture = anim->texture;
			if (sf::VertexBuffer::isAvailable())
				target.draw(anim->buff[atFrame], states);
			else
				target.draw(anim->arr[atFrame], states);
		}
	};
}